/*
 * Copyright (C) 2025 Phosh.mobi e.V.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <glib.h>

G_BEGIN_DECLS

#include "ptk-display.h"
#include "ptk-types.h"
#include "ptk-subsurface.h"
#include "ptk-surface.h"
#include "ptk-toplevel.h"

G_END_DECLS
