/*
 * Copyright (C) 2025 Phosh.mobi e.V.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include "ptk.h"

G_BEGIN_DECLS

bool                    ptk_demo_cutouts                         (PtkDisplay *display);
bool                    ptk_demo_subsurface                      (PtkDisplay *display);

G_END_DECLS
